#! /usr/bin/env python3

"""
N-fold pendulum
author: Harald von Fellenberg, hvf@hvf.ch
website: https://www.hvf.ch
License: BSD
based on Lagrange formalism 
17.11.2019 08.12.2019

adapted from code at http://matplotlib.sourceforge.net/examples/animation/double_pendulum_animated.py

Double pendulum formula translated from the C code at
http://www.physics.usyd.edu.au/~wheat/dpend_html/solve_dpend.c

author: Jake Vanderplas
email: vanderplas@astro.washington.edu
website: http://jakevdp.github.com
license: BSD
Please feel free to use and modify this, but keep the above information. Thanks!
hvf 23.01.2018 - add Epot, Ekin to graphics
hvf 21.11.2019 - read parameters from command line
hvf 08.12.2019 - extend to arbitrary number of pendulums
hvf 08.12.2019 - add title to graphic 
hvf 20.12.2019 - scale size of display to total length of arms

Usage:
    python Nfold-pendulum.py [N L1 .. LN M1 .. MN Phi1 .. PhiN Omega1 .. OmegaN]
    Parameters:
        N   number of pendulum rods
        L1 .. LN    length of rods 1 .. N
        M1 .. MN    mass of weight 1 .. N
        Phi1 .. PhiN    initial angle of rods in degrees
        Omega1 .. OmegaN    initial angular velocity of rods in degrees/sec

    Default values: 2   2 1     1 2     180 -90     0 0
                    N   L       M       Phi         Omega
"""

import numpy as np
import matplotlib.pyplot as plt
import scipy.integrate as integrate
import matplotlib.animation as animation
import sys

class Nfold_Pendulum:
    """Nfold_Pendulum Class

    init_state is N, Larr, Marr, Phiarr, Omegaarr
    where N is the number of pendula, Larr the list of pendulum lengths [m],
    Marr the list of pendulum masses [kg], Phiarr the list of initial
    angular positions in degrees, Omegaarr the list of angular velocities in
    degrees/sec.
    G is the acceleration of gravitiy, taken as 10 m/sec2.
    """
    def __init__(self,
                N=2,
                Larr=[2.0, 1.0],
                Marr=[2.0, 1.0],
                Phiarr=[180, -90],
                Omegaarr=[0, 0]):
        # set instance variables
        self.N = N
        self.L = np.asarray(Larr, dtype='float') 
        self.M = np.asarray(Marr, dtype='float') 
        # convert angles to radians
        self.Phi = np.radians(np.asarray(Phiarr, dtype='float'))
        self.Omega = np.radians(np.asarray(Omegaarr, dtype='float'))
        self.G = 10.0     # acceleration due to gravity, in m/s^2
        self.origin = (0, 0)
        self.time_elapsed = 0

        # state is Phi, Omega flattened in the state array
        self.state = np.asarray([self.Phi, self.Omega]).flatten()
    
        # additional stuff: Mk = sum of M[k:]
        self.Mk = np.zeros(N)
        for k in range(N):
            self.Mk[k] = np.sum(self.M[k:])
        
        self.params = (self.N, self.L, self.M, self.Mk,
                       self.G, self.origin)

        # print params
        print ("Initial Parameters")
        print ("N     = ", self.N)
        print ("L     = ", self.L)
        print ("M     = ", self.M)
        print ("Mk    = ", self.Mk)
        print ("G     = ", self.G)
        print ("Phi   = ", self.Phi)
        print ("Omega = ", self.Omega)
        print ("state (Phi[0:N], Omega[0:N]) = ", self.state)

    def getParams(self):
        (N, L, M, Mk, G, origin) = self.params
        return (N, L, M, Mk, G, origin)

    def position(self):
        """compute the current x,y positions of the pendulum arms"""
        (N, L, M, Mk, G, origin) = self.params
        #print ("Position: state ", self.state[0:N])
        Phi = np.array(np.copy(self.state[0:N]))
        Omega = np.array(np.copy(self.state[N:])) # not used :-)
        x = np.ndarray(N+1)
        y = np.ndarray(N+1)
        x[0] = origin[0]
        y[0] = origin[1]
        x[1:] = L*np.sin(Phi)
        y[1:] = -L*np.cos(Phi)
        x = np.cumsum(x)
        y = np.cumsum(y)

        #print ("Position Phi ", Phi)
        #print ("Position Omega ", Omega)
        #print ("Position x ", x)
        #print ("Position y ", y)
        return (x, y)

    def energy(self):
        """compute the energy of the current state"""
        (N, L, M, Mk, G, origin) = self.params

        Phi = self.state[:N]
        Omega = self.state[N:]
        x = np.cumsum([L*np.sin(Phi)]) # not used
        y = np.cumsum([-L*np.cos(Phi)])

        vx = np.cumsum([L*Omega*np.cos(Phi)])
        vy = np.cumsum([L*Omega*np.sin(Phi)])

        U = G * np.sum(M * y)
        K = 0.5 * np.sum(M * (vx**2 + vy**2))
        #print ("Total Energy, Pot, Kin ", U+K, U, K)
        return (U + K, U, K)

    def dstate_dt(self, state, t):
        """compute the derivative of the given state"""
        (N, L, M, Mk, G, origin) = self.params

        verbose = False

        # take phi, omega out of state
        phi = state[0:N]
        omega = state [N:]

        # NxN matrix
        A = np.zeros((N, N))
        R = np.zeros((N, N))

        # omegasquare vector
        omegasq = omega**2

        # gravity vector
        Gforce = -G * Mk * np.sin(phi)

        # Key Part
        # j is row, k column
        for k in range(N):
            A[k, k] = Mk[k] * L[k]
            for j in range(k):  # from 0 to k-1, upper right triangle
                A[j, k] = Mk[k] * L[j] * np.cos(phi[j] - phi[k])
                R[j, k] = Mk[k] * L[j] * np.sin(phi[j] - phi[k])
            for j in range(k+1, N): # from k+1 to N-1, lower left triangle
                A[j, k] = Mk[j] * L[j] * np.cos(phi[j] - phi[k])
                R[j, k] = Mk[j] * L[j] * np.sin(phi[j] - phi[k])

        RR = np.dot(omegasq, R) + Gforce

        Ainv = np.linalg.inv(A)

        # derivation vector omega prime ( phi prime prime)
        omegaprime = np.dot(RR, Ainv)
        d = np.concatenate((omega, omegaprime))
        # print if requested
        if verbose:
            print ("A = \n", A)
            print ("Det(A) = ", np.linalg.det(A))
            print ("Ainv = \n", Ainv)
            print ("OmegaDot = RR * Ainv\n", omegaprime, omegaprime.shape)
            # omega is phi prime
            print ("Omega = ", omega, omega.shape)
            # concatenate phi prime and omega prime to go into the ODE 
            print("Deriv = ", d, d.shape)

        return d
        
    def step(self, dt):
        #print ("Step: dt = ", dt)
        #print ("Starting state = ", self.state)
        """execute one time step of length dt and update state"""
        self.state = integrate.odeint(self.dstate_dt, self.state, [0, dt], full_output=0)[1]
        self.time_elapsed += dt
        #print ("Final state = ", self.state)

#------------------------------------------------------------
def Version():
    print ("N-fold pendulum Version 08.12.2019 by Harald von Fellenberg")

def Usage():
    Version()
    print ("""
Usage:
    python Nfold-pendulum.py [N L1 .. LN M1 .. MN Phi1 .. PhiN Omega1 .. OmegaN]
    Parameters:
        N   number of pendulum rods
        L1 .. LN    length of rods 1 .. N
        M1 .. MN    mass of weight 1 .. N
        Phi1 .. PhiN    initial angle of rods in degrees
        Omega1 .. OmegaN    initial angular velocity of rods in degrees/sec

    Default values: 2   2 1     1 2     180 -90     0 0
                    N   L       M       Phi         Omega""")
    sys.exit(1)
#
# Read command line parameters
Larr = []       # list of lengths
Marr = []       # list of masses
Phiarr =  []    # list of initial angles (degrees)
Omegaarr = []   # list of initial angular velocities (degrees/sec)

if (len(sys.argv) > 1): # parameter list not empty
    arg = sys.argv[1]
    if arg == '-h' or arg == '--help': Usage()
    N = int(arg)
    if (len(sys.argv) < 4*N+2 or N<1): # not enough parameters
        Usage()
    for i in range(N):  
        Larr.append(sys.argv[2+i])              # length
        Marr.append(sys.argv[2+N+i])            # mass
        Phiarr.append(sys.argv[2+2*N+i])        # angle phi in degrees
        Omegaarr.append(sys.argv[2+3*N+i])      # angle velocity in degrees/sec
        
else:   # no parameters: use standard values for N=2
    N=2
    Larr=[2.0, 1.0]
    Marr=[1.0, 2.0]
    Phiarr=[180., -90.]
    Omegaarr=[0, 0]
# check parameters
if float(min(Marr)) <= 0.0:
    print ("Masses must be positive"); Usage()
if float(min(Larr)) <= 0.0:
    print ("Lengths must be positive"); Usage()
Version()

# set up initial state and global variables
pendulum = Nfold_Pendulum(N, Larr, Marr, Phiarr, Omegaarr)

# time step
dt = 1./30 # 30 fps

#------------------------------------------------------------
# set up figure and animation
# set xlim, ylim to sum of Larr
L = np.asarray(Larr, dtype='float') 
Lsum = max(sum(L), 4)
fig = plt.figure()
ax = fig.add_subplot(111, aspect='equal', autoscale_on=False,
                     xlim=(-Lsum, Lsum), ylim=(-Lsum, Lsum))
                     #xlim=(-4, 4), ylim=(-4, 4))
ax.grid()

(N, L, M, Mk, G, origin) = pendulum.getParams()
lentext = ""
masstext = ""
for i in range(N):
    Ltext = "{0:.1f}, ".format(L[i])
    lentext = "".join([lentext, Ltext])
    Mtext = "{0:.1f}, ".format(M[i])
    masstext = "".join([masstext, Mtext])
    
print (lentext)
print (masstext)
fig.suptitle("Nfold Pendulum\nN= %d, Arms = %s Masses = %s" % (N, lentext, masstext))

line, = ax.plot([], [], 'o-', lw=2)
time_text = ax.text(0.02, 0.95, '', transform=ax.transAxes)
energy_text = ax.text(0.02, 0.90, '', transform=ax.transAxes)
pot_energy_text = ax.text(0.02, 0.85, '', transform=ax.transAxes)
kin_energy_text = ax.text(0.02, 0.80, '', transform=ax.transAxes)

def init():
    """initialize animation"""
    line.set_data([], [])
    time_text.set_text('')
    energy_text.set_text('')
    pot_energy_text.set_text('')
    kin_energy_text.set_text('')
    return line, time_text, energy_text

def animate(i):
    """perform animation step"""
    global pendulum, dt
    pendulum.step(dt)
    
    line.set_data(*pendulum.position())
    time_text.set_text('time = %.1f' % pendulum.time_elapsed)
    (Etot, Epot, Ekin) = pendulum.energy()
    energy_text.set_text('Tot energy = %.3f J' % Etot)
    pot_energy_text.set_text('Pot energy = %.3f J' % Epot)
    kin_energy_text.set_text('Kin energy = %.3f J' % Ekin)
    return line, time_text, energy_text, pot_energy_text, kin_energy_text

# choose the interval based on dt and the time to animate one step
from time import time
t0 = time()
animate(0)
t1 = time()
interval = 1000 * dt - (t1 - t0)

print ("Interval [msec] = %.2f" % interval)

# 450 frames willl create a film of 15 seconds.
ani = animation.FuncAnimation(fig, animate, frames=450,
              interval=interval, blit=True, init_func=init)

# save the animation as an mp4.  This requires ffmpeg or mencoder to be
# installed.  The extra_args ensure that the x264 codec is used, so that
# the video can be embedded in html5.  You may need to adjust this for
# your system: for more information, see
# http://matplotlib.sourceforge.net/api/animation_api.html
# un-comment the next 3 lines to create a 15 seconds film of the movements.
##film_file = "Nfold_pendulum_" + str(N) + ".mp4"
##print ("Film : ", film_file)
##ani.save(film_file, fps=30, extra_args=['-vcodec', 'libx264'])

plt.show()
